require 'English'
require 'rake'

module GzRelease
  # Rake tasks for building a docker image
  class Tasks
    include Rake::DSL

    DOCKERFILE_PATH = File.expand_path('./Dockerfile')
    DOCKER_CACHE_PATH = File.expand_path('~/.docker')
    DOCKER_IMAGE_PATH = File.join(DOCKER_CACHE_PATH, 'image.tar')
    TIMESTAMP_FORMAT = '%Y_%m_%d_%H_%M_%S'

    BuildError = Class.new(StandardError)

    class << self
      def install(*args)
        new(*args).install
      end
    end

    attr_reader :version

    def initialize(options = {})
      @image_name, @version = options.values_at(:image_name, :version)

      raise BuildError, 'Image name not specified.' unless image_name
    end

    def install
      namespace :release do
        desc 'Creates a latest tag and version tag if available. Version is set in Rakefile'

        desc 'Creates a git tag and image tag of the current timestamp.'
        task(:tag_timestamp){tag_timestamp}

        desc 'Pulls latest version of the image from dockerhub'
        task(:pull){pull}

        desc 'Loads ~/.docker/image.tar into docker cache. If file does not exist, pulls from dockerhub.'
        task(:load){load}

        desc 'Dumps docker cache into ~/.docker/image.tar'
        task(:dump){dump}

        desc 'Builds the docker image from generated Dockerfile'
        task(:build){build}

        namespace :tag do
          desc 'Creates a tag for the branch for the docker image'
          task(:branch){tag_branch}

          desc 'Creates a tag for version specified in Rakefile for docker image'
          task(:version){tag_version}

          desc 'Creates latest tag for the docker image'
          task(:latest){tag_latest}

          desc 'Creates release tag for the docker image'
          task(:release){tag_release}

          desc 'Creates timestamp tag for the docker image and git repo'
          task(:timestamp){tag_timestamp}
        end

        namespace :push do
          desc 'Pushes docker image tagged as current git branch.'
          task(branch: :'release:tag:branch'){push_branch}

          desc 'Pushes docker image tagged as version specified in Rakefile'
          task(version: :'release:tag:version'){push_version}

          desc 'Pushes docker image tagged as latest'
          task(latest: :'release:tag:latest'){push_latest}

          desc 'Pushes docker image tagged as release'
          task(release: :'release:tag:release'){push_release}

          desc 'Pushes docker image and git tag tagged by the timestamp.'
          task(timestamp: :'release:tag:timestamp'){push_timestamp}
        end
      end
    end

    def pull
      run("docker pull #{image_version}")
    rescue BuildError
      puts 'Skipping pull. Note, this should generally only happen on first build.'
    end

    def load
      if File.exist?(DOCKER_IMAGE_PATH)
        run("docker load -i #{DOCKER_IMAGE_PATH}")
      else
        pull
      end
    end

    def dump
      require 'fileutils'

      FileUtils.mkdir_p(DOCKER_CACHE_PATH)
      run("docker save #{image_latest} > #{DOCKER_IMAGE_PATH}")
    end

    def build
      run("docker build -t #{image_sha} .")
    end

    def tag(tag_name)
      run("docker tag -f #{image_sha} #{tag_name}")
    end

    def git_tag(tag_name)
      run("git tag -a #{tag_name} -m 'Tagging #{tag_name} from #{branch}(#{sha})' #{sha}")
    end

    def push(tag_name)
      login

      run("docker push #{tag_name}")
    end

    def git_push(tag_name)
      run("git push origin #{tag_name}")
    end

    def login
      email = ENV['DOCKER_EMAIL']
      user = ENV['DOCKER_USER']
      password = ENV['DOCKER_PASSWORD']

      errors = []
      errors << 'Missing DOCKER_EMAIL environment variable.' unless email
      errors << 'Missing DOCKER_USER environment variable.' unless user
      errors << 'Missing DOCKER_PASSWORD environment variable.' unless password
      raise BuildError, "Encountered the following errors:\n #{errors.join("\n")}" if errors.any?

      run("docker login -e #{email} -u #{user} -p #{password} #{docker_repository}", display_command: false)
    end

    def run(command, options = {display_command: true})
      puts command if options[:display_command]
      system(command)

      raise BuildError, 'There was a problem executing this command.' unless $CHILD_STATUS == 0
    end

    def image_name
      name = ''
      name << "#{docker_repository}/" unless docker_repository.empty?
      name << @image_name
    end

    def docker_repository
      ENV['DOCKER_REPOSITORY'] || ''
    end

    # SHA
    def image_sha
      "#{image_name}:#{sha}"
    end

    def sha
      @sha ||= `git rev-parse HEAD`.strip
    end

    # Branch
    def tag_branch
      tag(image_branch)
    end

    def push_branch
      push(image_branch)
    end

    def image_branch
      "#{image_name}:#{branch}"
    end

    def branch
      @branch ||= (ENV['CIRCLE_BRANCH'] ? ENV['CIRCLE_BRANCH'] : `git rev-parse --abbrev-ref HEAD`).strip.gsub(/\//,'_')
    end

    # Version
    def tag_version
      return unless version

      tag(image_version)
    end

    def push_version
      return unless version

      push(image_version)
    end

    def image_version
      "#{image_name}:#{version}"
    end

    # Latest
    def tag_latest
      tag(image_latest)
    end

    def push_latest
      push(image_latest)
    end

    def image_latest
      "#{image_name}:latest"
    end

    # Release
    def tag_release
      tag(image_release)
    end

    def push_release
      push(image_release)
    end

    def image_release
      "#{image_name}:release"
    end

    # Timestamp
    def tag_timestamp
      git_tag(timestamp)
      tag(image_timestamp)
    end

    def push_timestamp
      git_push(timestamp)
      push(image_timestamp)
    end

    def image_timestamp
      "#{image_name}:#{timestamp}"
    end

    def timestamp
      @timestamp ||= Time.now.utc.strftime(TIMESTAMP_FORMAT)
    end
  end
end
